/*
 * Copyright 1993 Network Computing Devices, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name Network Computing Devices, Inc. not be
 * used in advertising or publicity pertaining to distribution of this
 * software without specific, written prior permission.
 *
 * THIS SOFTWARE IS PROVIDED 'AS-IS'.  NETWORK COMPUTING DEVICES, INC.,
 * DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NONINFRINGEMENT.  IN NO EVENT SHALL NETWORK
 * COMPUTING DEVICES, INC., BE LIABLE FOR ANY DAMAGES WHATSOEVER, INCLUDING
 * SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, INCLUDING LOSS OF USE, DATA,
 * OR PROFITS, EVEN IF ADVISED OF THE POSSIBILITY THEREOF, AND REGARDLESS OF
 * WHETHER IN AN ACTION IN CONTRACT, TORT OR NEGLIGENCE, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * $NCDId: @(#)snd.c,v 1.14 1993/03/02 22:40:44 lemke Exp $
 */

#include <stdio.h>
#include <strings.h>
#include <audio/snd.h>
#include <audio/fileutil.h>

extern char    *malloc();

#define Err() { if (fp != stdin) fclose(fp); return NULL; }

FILE           *
SndOpenFileForReading(name, hdr, commentp)
_SndConst char *name;
SndHeader      *hdr;
char          **commentp;
{
    FILE           *fp;
    int	endian = 1;

    if (!(fp = strcmp(name, "-") ? fopen(name, "r") : stdin))
	return NULL;

    if (fread(hdr, 1, sizeof(SndHeader), fp) != sizeof(SndHeader))
	Err();

    if (*(char *)&endian) {	/* header is big-endian */
	char	n;

	swapl(&hdr->magic, n);
	swapl(&hdr->dataOffset, n);
	swapl(&hdr->dataSize, n);
	swapl(&hdr->format, n);
	swapl(&hdr->sampleRate, n);
	swapl(&hdr->tracks, n);
    }

    if (hdr->magic != SND_MAGIC)
	Err();

    if (commentp)
    {
	int             size = hdr->dataOffset - sizeof(SndHeader);

	if (size)
	{
	    if (!(*commentp = (char *) malloc(size + 1)))
		Err();

	    if (fread(*commentp, 1, size, fp) != size)
	    {
		free(*commentp);
		Err();
	    }

	    *(*commentp + size) = 0;
	}
	else
	    *commentp = FileCommentFromFilename(name);
    }

    if (fp != stdin && hdr->dataSize == SND_DATA_SIZE_UNKNOWN)
    {
	fseek(fp, 0, 2);
	hdr->dataSize = ftell(fp) - hdr->dataOffset;
	fseek(fp, hdr->dataOffset, 0);
    }

    return fp;
}

FILE           *
SndOpenFileForWriting(name, hdr, comment)
_SndConst char *name;
SndHeader      *hdr;
_SndConst char *comment;
{
    FILE           *fp;
    int             commentLen;
    int             is_stdio = 0;

    commentLen = comment ? strlen(comment) : 0;
    hdr->magic = SND_MAGIC;
    hdr->dataSize = SND_DATA_SIZE_UNKNOWN;
    hdr->dataOffset = sizeof(SndHeader) + commentLen;

    if (strcmp(name, "-") == 0)
    {
	fp = stdout;
	is_stdio = 1;
    }
    else
    {
	fp = fopen(name, "w");
    }
    if (fp)
    {
	int	endian = 1;

	if (*(char *)&endian) {	/* header is big-endian */
	    char	n;

	    swapl(&hdr->magic, n);
	    swapl(&hdr->dataOffset, n);
	    swapl(&hdr->dataSize, n);
	    swapl(&hdr->format, n);
	    swapl(&hdr->sampleRate, n);
	    swapl(&hdr->tracks, n);
	}

	if (fwrite(hdr, 1, sizeof(SndHeader), fp) == sizeof(SndHeader))
	    if (!commentLen ||
		fwrite(comment, 1, commentLen, fp) == commentLen)
		return fp;

	if (!is_stdio)
	    fclose(fp);
    }

    return NULL;
}

unsigned int
SndSizeofFormat(format)
int             format;
{
    static unsigned int f2bps[] =
    {
	0,					/* unused */
	1,					/* SND_FORMAT_ULAW_8 */
	1,					/* SND_FORMAT_LINEAR_8 */
	2,					/* SND_FORMAT_LINEAR_16 */
    };

    if (format < 0 || format > (sizeof(f2bps) / sizeof(f2bps[0]) - 1))
	format = 0;

    return f2bps[format];
}

static struct
{
    int             format;
    char           *string;
}               formats[] =
{
    SND_FORMAT_ULAW_8, "8-bit ulaw",
    SND_FORMAT_LINEAR_8, "8-bit linear",
    SND_FORMAT_LINEAR_16, "16-bit linear",
    SND_FORMAT_LINEAR_24, "24-bit linear",
    SND_FORMAT_LINEAR_32, "32-bit linear",
    SND_FORMAT_FLOAT_32, "32-bit IEEE floating point",
    SND_FORMAT_FLOAT_64, "64-bit IEEE floating point",
    SND_FORMAT_ULAW_COMPRESSED_8,
    "8-bit ulaw compressed using the CCITT G.721 ADPCM voice data format scheme"
};

_SndConst char *
SndFormatToString(format)
unsigned int    format;
{
    int             i;

    for (i = 0; i < sizeof(formats) / sizeof(formats[0]); i++)
	if (formats[i].format == format)
	    return formats[i].string;

    return "Unknown";
}

int
SndStringToFormat(t)
_SndConst char *t;
{
    int             i;

    for (i = 0; i < sizeof(formats) / sizeof(formats[0]); i++)
	if (!strcasecmp(t, formats[i].string))
	    return formats[i].format;

    return -1;
}
