/*
 * Copyright 1993 Network Computing Devices, Inc.
 * 
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name Network Computing Devices, Inc. not be
 * used in advertising or publicity pertaining to distribution of this
 * software without specific, written prior permission.
 * 
 * THIS SOFTWARE IS PROVIDED 'AS-IS'.  NETWORK COMPUTING DEVICES, INC.,
 * DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT
 * LIMITATION ALL IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NONINFRINGEMENT.  IN NO EVENT SHALL NETWORK
 * COMPUTING DEVICES, INC., BE LIABLE FOR ANY DAMAGES WHATSOEVER, INCLUDING
 * SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, INCLUDING LOSS OF USE, DATA,
 * OR PROFITS, EVEN IF ADVISED OF THE POSSIBILITY THEREOF, AND REGARDLESS OF
 * WHETHER IN AN ACTION IN CONTRACT, TORT OR NEGLIGENCE, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 * $NCDId: @(#)soundlib.c,v 1.22 1993/03/15 22:48:53 greg Exp $
 */

#include <stdio.h>
#include <string.h>				/* or <strings.h>... */
#include "Alibint.h"
#include "soundlib.h"

unsigned int    AuSoundFileChunkSize = 10 * 1024,	/* in bytes */
                AuSoundPortDuration = 10,	/* in seconds */
                AuSoundPortLowWaterMark = 25,	/* in percent */
                AuSoundPortHighWaterMark = 75;	/* in percent */

AuBucketID
AuSoundCreateBucketFromFile(aud, filename, access, ret_attr, ret_status)
AuServer       *aud;
_SoundConst char *filename;
unsigned long   access;
AuBucketAttributes **ret_attr;			/* RETURN */
AuStatus       *ret_status;
{
    SoundHeader     hdr;
    FILE           *fp;
    AuBucketID      bucket;
    AuFlowID        flow;
    int             import;
    char           *description,
                   *buf;
    unsigned int    count;
    AuString        desc;
    AuBool          done = AuFalse;

    if (!(fp = SoundOpenFileForReading(filename, &hdr, &description)))
	return AuNone;

    if (!(buf = (char *) Aumalloc(AuSoundFileChunkSize)))
    {
	fclose(fp);
	return AuNone;
    }

    /* seek to the start of the data */
    /* SUPPRESS 558 */
    SoundSeek(fp, &hdr, 0, 0);

    desc.type = AuStringLatin1;
    desc.len = strlen(description);
    desc.data = description;

    bucket = AuCreateBucket(aud, hdr.format, hdr.tracks, access, hdr.sampleRate,
			    hdr.dataSize / hdr.tracks /
			  SoundSizeofFormat(hdr.format), &desc, ret_status);

    if (bucket)
    {
	/*
	 * Get an unclocked flow into a bucket.  We can do this operation
	 * synchronously, so we don't have to horse around with events.
	 */
	flow = AuGetScratchFlowToBucket(aud, bucket, &import, ret_status);

	if (flow)
	{
	    while (!done)
	    {
		count = aumin(AuSoundFileChunkSize, hdr.dataSize);
		count = fread(buf, 1, count, fp);
		hdr.dataSize -= count;
		done = !(count && hdr.dataSize);

		(void) AuWriteElement(aud, flow, import, count, buf,
				      done, ret_status);
	    }

	    AuReleaseScratchFlow(aud, flow, ret_status);
	}

	if (ret_attr)
	    *ret_attr = AuGetBucketAttributes(aud, bucket, ret_status);
    }

    Aufree(buf);
    free(description);
    fclose(fp);

    return bucket;
}

AuBool
AuSoundCreateFileFromBucket(aud, filename, type, bucket, ret_status)
AuServer       *aud;
_AuConst char  *filename;
int             type;
AuBucketID      bucket;
AuStatus       *ret_status;
{
    AuBucketAttributes *ba;
    AuFlowID        flow;
    int             export;
    SoundHeader     hdr;
    AuBool          result = AuTrue;
    char           *p;
    FILE           *fp;
    unsigned long   size;

    if (!(ba = AuGetBucketAttributes(aud, bucket, ret_status)))
	return AuFalse;

    size = AuBucketNumSamples(ba) * SoundSizeofFormat(AuBucketFormat(ba)) *
	AuBucketNumTracks(ba);

    if (!(p = (char *) Aumalloc(AuSoundFileChunkSize)))
    {
	AuFreeBucketAttributes(aud, 1, ba);
	return AuFalse;
    }

    hdr.format = AuBucketFormat(ba);
    hdr.sampleRate = AuBucketSampleRate(ba);
    hdr.tracks = AuBucketNumTracks(ba);
    hdr.sound_type = type;

    /* For now, assume that description is ISOLatin1 */

    if (!(fp = SoundOpenFileForWriting(filename, &hdr,
				       AuBucketDescription(ba)->data)))
    {
	Aufree(p);
	AuFreeBucketAttributes(aud, 1, ba);
	return AuFalse;
    }

    /*
     * Get an unclocked flow from a bucket.  We can do this operation
     * synchronously, so we don't have to horse around with events.
     */
    flow = AuGetScratchFlowFromBucket(aud, bucket, &export, ret_status);

    if (flow)
    {
	unsigned long   wanted = size,
	                got = 0;

	while (wanted &&
	       (got = AuReadElement(aud, flow, export,
				    aumin(wanted, AuSoundFileChunkSize), p,
				    ret_status)))
	{
	    if (fwrite(p, 1, got, fp) != got)
	    {
		result = AuFalse;
		break;
	    }

	    wanted -= got;
	}

	AuReleaseScratchFlow(aud, flow, ret_status);
    }

    free(p);
    AuFreeBucketAttributes(aud, 1, ba);
    fclose(fp);

    return result;
}

AuBucketID
AuSoundCreateBucketFromData(aud, hdr, data, description, access, ret_attr, ret_status)
AuServer       *aud;
SoundHeader    *hdr;
AuPointer       data;
_AuConst char  *description;
unsigned long   access;
AuBucketAttributes **ret_attr;			/* RETURN */
AuStatus       *ret_status;
{
    AuBucketID      bucket;
    AuFlowID        flow;
    int             import;
    AuString        desc;

    if (!hdr)
    {
	char           *t = (char *) data;

	hdr = (SoundHeader *) data;
	t += hdr->dataOffset;
	data = (AuPointer) t;
    }
    if (hdr->dataSize == SND_DATA_SIZE_UNKNOWN)
    {
	return AuNone;
    }
    desc.type = AuStringLatin1;
    desc.len = strlen(description);
    desc.data = (char *) description;

    bucket = AuCreateBucket(aud, hdr->format, hdr->tracks, access,
			    hdr->sampleRate,
			    hdr->dataSize / hdr->tracks /
			    SoundSizeofFormat(hdr->format),
			    &desc, ret_status);
    if (bucket)
    {
	/*
	 * Get an unclocked flow into a bucket.  We can do this operation
	 * synchronously, so we don't have to horse around with events.
	 */
	flow = AuGetScratchFlowToBucket(aud, bucket, &import, ret_status);

	if (flow)
	{
	    (void) AuWriteElement(aud, flow, import, hdr->dataSize, data,
				  AuTrue, ret_status);
	    AuReleaseScratchFlow(aud, flow, ret_status);
	}
	if (ret_attr)
	    *ret_attr = AuGetBucketAttributes(aud, bucket, ret_status);

    }
    return bucket;
}

AuPointer
AuSoundCreateDataFromBucket(aud, bucket, pHeadPtr, pComment, ret_status)
AuServer       *aud;
AuBucketID      bucket;
SoundHeader    *pHeadPtr;
char          **pComment;
AuStatus       *ret_status;
{
    AuBucketAttributes *ba;
    AuFlowID        flow;
    int             export;
    char           *p;
    char           *comment = NULL;
    long            nbytes;
    int             clen;

    *pComment = NULL;
    bzero((char *) pHeadPtr, sizeof(SoundHeader));

    ba = AuGetBucketAttributes(aud, bucket, ret_status);
    if (!ba)
	return NULL;

    nbytes = AuBucketNumSamples(ba) * SoundSizeofFormat(AuBucketFormat(ba)) *
	AuBucketNumTracks(ba);

    p = (char *) Aumalloc(nbytes);
    clen = AuBucketDescription(ba)->len + 1;
    comment = (char *) Aumalloc(clen);

    if (!p || !comment)
    {
	AuFreeBucketAttributes(aud, 1, ba);
	if (p)
	    Aufree(p);
	if (comment)
	    Aufree(comment);
	return NULL;
    }
    pHeadPtr->sound_type = SND_MAGIC;	/* XXX is snd the proper default? */
    pHeadPtr->dataSize = nbytes;
    /* XXX -- hack -- its really not there at all... */
    pHeadPtr->dataOffset = sizeof(SoundHeader) + clen;
    pHeadPtr->format = AuBucketFormat(ba);
    pHeadPtr->sampleRate = AuBucketSampleRate(ba);
    pHeadPtr->tracks = AuBucketNumTracks(ba);
    pHeadPtr->bitsPerSample = SoundSizeofFormat(AuBucketFormat(ba));

    /*
     * Get an unclocked flow from a bucket.  We can do this operation
     * synchronously, so we don't have to horse around with events.
     */
    flow = AuGetScratchFlowFromBucket(aud, bucket, &export, ret_status);
    if (flow)
    {
	AuReadElement(aud, flow, export, nbytes, p, ret_status);
	AuReleaseScratchFlow(aud, flow, ret_status);
    }
    AuFreeBucketAttributes(aud, 1, ba);

    strncpy(comment, AuBucketDescription(ba)->data, clen);
    comment[clen] = '\0';
    *pComment = comment;
    return (AuPointer) p;
}

typedef struct
{
    FILE           *fp;
    SoundHeader     hdr;
    AuFlowID        flow;
    char           *buf;
    AuPointer       callback_data;
    int             loopCount;
    void            (*callback) (),
                    (*dataHandler) (),
                    (*dataHandlerStop) ();
}               PrivRec, *PrivPtr;

static void
sendData(aud, priv, numBytes)
AuServer       *aud;
PrivPtr         priv;
unsigned long   numBytes;
{
    int             n;

    if ((n = fread(priv->buf, 1, aumin(numBytes, priv->hdr.dataSize),
		   priv->fp)) > 0)
    {
	AuWriteElement(aud, priv->flow, 0, n, priv->buf, n != numBytes, NULL);
	priv->hdr.dataSize -= n;
    }
}

static void
receiveData(aud, priv, numBytes)
AuServer       *aud;
PrivPtr         priv;
unsigned long   numBytes;
{
    unsigned long   n;

    n = AuReadElement(aud, priv->flow, 1, numBytes, priv->buf, NULL);
    fwrite(priv->buf, 1, n, priv->fp);
}

static          AuBool
EventHandler(aud, ev, handler)
AuServer       *aud;
AuEvent        *ev;
AuEventHandlerRec *handler;
{
    PrivPtr         priv = (PrivPtr) handler->data;

    switch (ev->type)
    {
	case AuEventTypeMonitorNotify:
	    if (priv->callback)
		(*priv->callback) (aud, handler, ev, priv->callback_data);
	    break;
	case AuEventTypeElementNotify:
	    {
		AuElementNotifyEvent *event = (AuElementNotifyEvent *) ev;

		switch (event->kind)
		{
		    case AuElementNotifyKindHighWater:
		    case AuElementNotifyKindLowWater:
			(*priv->dataHandler) (aud, priv, event->num_bytes);
			break;
		    case AuElementNotifyKindState:
			switch (event->cur_state)
			{
			    case AuStateStop:
				if (priv->dataHandlerStop)
				    (*priv->dataHandlerStop) (aud, priv,
							  event->num_bytes);

				if (priv->callback)
				    (*priv->callback) (aud, handler, ev,
						       priv->callback_data);

				if (priv->loopCount &&
				    event->reason == AuReasonEOF)
				{
				    if (!--priv->loopCount)
					AuStopFlow(aud, priv->flow, NULL);
				    break;
				}

				AuUnregisterEventHandler(aud, handler);
				AuReleaseScratchFlow(aud, priv->flow, NULL);
				if (priv->fp)
				    SoundCloseFile(priv->fp);
				Aufree(priv);
				break;
			    case AuStatePause:
				(*priv->dataHandler) (aud, priv,
						      event->num_bytes);
				break;
			}
			break;
		}
	    }
    }

    return AuTrue;
}

#define ABORT()								       \
{									       \
    SoundCloseFile(priv->fp);						       \
    Aufree(priv);							       \
									       \
    return NULL;							       \
}

/* play a Sound file */
AuEventHandlerRec *
AuSoundPlayFromFile(aud, filename, device, volume, callback, callback_data,
		    ret_flow, ret_mult_elem, ret_mon_elem, ret_status)
AuServer       *aud;
_AuConst char  *filename;
AuDeviceID      device;
AuFixedPoint    volume;
void            (*callback) ();
AuPointer       callback_data;
AuFlowID       *ret_flow;
int            *ret_mult_elem,
               *ret_mon_elem;
AuStatus       *ret_status;
{
    AuElement       elements[4];
    AuElementState  states[1];
    PrivPtr         priv;
    unsigned int    importSize,
                    bufSize,
                    i;
    SoundHeader     hdr;
    FILE           *fp;
    AuEventHandlerRec *handler;

    /* open the file and get the header */
    if (!(fp = SoundOpenFileForReading(filename, &hdr, (char **) NULL)))
	return NULL;

    /* seek to the start of the data */
    /* SUPPRESS 558 */
    SoundSeek(fp, &hdr, 0, 0);

    /* calc the size of the import */
    importSize = hdr.sampleRate * AuSoundPortDuration;
    bufSize = importSize * hdr.tracks * SoundSizeofFormat(hdr.format);

    /* alloc private data */
    if (!(priv = (PrivPtr) Aumalloc(PAD4(sizeof(PrivRec)) + bufSize)))
    {
	SoundCloseFile(fp);
	return NULL;
    }

    priv->loopCount = 0;
    priv->callback = callback;
    priv->callback_data = callback_data;
    priv->dataHandler = sendData;
    priv->dataHandlerStop = NULL;
    priv->buf = ((char *) priv) + PAD4(sizeof(PrivRec));
    priv->fp = fp;
    priv->hdr = hdr;

    /* if no ouput device was specified, look for an appropriate one */
    if (device == AuNone)
    {
	for (i = 0; i < AuServerNumDevices(aud); i++)
	    if ((AuDeviceKind(AuServerDevice(aud, i)) ==
		 AuComponentKindPhysicalOutput) &&
		AuDeviceNumTracks(AuServerDevice(aud, i)) == hdr.tracks)
	    {
		device = AuDeviceIdentifier(AuServerDevice(aud, i));
		break;
	    }
	/* abort if we didn't find an appropriate device */
	if (device == AuNone)
	    ABORT();
    }
    /* create the flow */
    if ((priv->flow = AuGetScratchFlow(aud, NULL)) == AuNone)
	ABORT();

    AuMakeElementImportClient(&elements[0], hdr.sampleRate,
			      hdr.format,
			      hdr.tracks,
			      AuTrue,
			      importSize,
			      importSize * AuSoundPortLowWaterMark / 100,
			      0, NULL);
    AuMakeElementMultiplyConstant(&elements[1], 0, volume);
    AuMakeElementExportDevice(&elements[2], 1, device, hdr.sampleRate,
			      AuUnlimitedSamples, 0, NULL);

    if (ret_mon_elem)
    {
	extern int      AuMonitorRate,
	                AuMonitorFormat;

	i = 4;
	*ret_mon_elem = 3;
	AuMakeElementExportMonitor(&elements[3], 0, AuMonitorRate,
				   AuMonitorFormat, hdr.tracks);
    }
    else
	i = 3;

    /* set up the flow */
    AuSetElements(aud, priv->flow, AuTrue, i, elements, ret_status);

    /* set up the event handler */
    if (!(handler = AuRegisterEventHandler(aud, AuEventHandlerIDMask,
					   0, priv->flow, EventHandler,
					   (AuPointer) priv)))
    {
	AuReleaseScratchFlow(aud, priv->flow, ret_status);
	ABORT();
    }
    /* start up the components */
    AuMakeElementState(&states[0], priv->flow, AuElementAll, AuStateStart);
    AuSetElementStates(aud, 1, states, ret_status);

    if (ret_flow)
	*ret_flow = priv->flow;

    if (ret_mult_elem)
	*ret_mult_elem = 1;

    return handler;
}

/* ARGSUSED */
/* record a Sound file */
AuEventHandlerRec *
AuSoundRecordToFile(aud, filename, device, gain, callback, callback_data,
		 mode, type, comment, rate, format, ret_flow, ret_mult_elem,
		    ret_status)
AuServer       *aud;
_AuConst char  *filename;
AuDeviceID      device;
AuFixedPoint    gain;
void            (*callback) ();
AuPointer       callback_data;
int             mode,
                format;
unsigned long   type,
                rate;
char           *comment;
AuFlowID       *ret_flow;
int            *ret_mult_elem;
AuStatus       *ret_status;
{
    AuElement       elements[2];
    AuElementState  states[1];
    PrivPtr         priv;
    unsigned int    exportSize,
                    bufSize,
                    i;
    SoundHeader     hdr;
    FILE           *fp;
    AuEventHandlerRec *handler;
    AuDeviceAttributes *d = NULL;

    /* get the device attributes */
    for (i = 0; i < AuServerNumDevices(aud); i++)
	if (AuDeviceIdentifier(AuServerDevice(aud, i)) == device)
	{
	    d = AuServerDevice(aud, i);
	    break;
	}

    if (!d)
	return NULL;

    hdr.sound_type = type;
    hdr.format = format;
    hdr.tracks = AuDeviceNumTracks(d);
    hdr.sampleRate = rate;

    /* calc the size of the export */
    exportSize = hdr.sampleRate * AuSoundPortDuration;
    bufSize = exportSize * hdr.tracks * SoundSizeofFormat(hdr.format);

    /* alloc private data */
    if (!(priv = (PrivPtr) Aumalloc(PAD4(sizeof(PrivRec)) + bufSize)))
	return NULL;

    /* create the flow */
    if ((priv->flow = AuGetScratchFlow(aud, NULL)) == AuNone)
    {
	Aufree(priv);
	return NULL;
    }

    AuMakeElementImportDevice(&elements[0], hdr.sampleRate, device,
			      AuUnlimitedSamples, 0, NULL);
    AuMakeElementExportClient(&elements[1], 0, hdr.sampleRate,
			      hdr.format,
			      hdr.tracks,
			      AuTrue,
			      exportSize,
			      exportSize * AuSoundPortHighWaterMark / 100,
			      0, NULL);

    /* set up the flow */
    AuSetElements(aud, priv->flow, AuTrue, 2, elements, ret_status);

    /* set up the event handler */
    if (!(handler = AuRegisterEventHandler(aud, AuEventHandlerTypeMask |
					   AuEventHandlerIDMask,
					   AuEventTypeElementNotify,
					   priv->flow, EventHandler, priv)))
    {
	AuReleaseScratchFlow(aud, priv->flow, ret_status);
	Aufree(priv);
	return NULL;
    }

    /* create the file */
    if (!(fp = SoundOpenFileForWriting(filename, &hdr, comment)))
	return NULL;

    priv->loopCount = 0;
    priv->fp = fp;
    priv->hdr = hdr;
    priv->buf = ((char *) priv) + PAD4(sizeof(PrivRec));
    priv->callback = callback;
    priv->callback_data = callback_data;
    priv->dataHandler = priv->dataHandlerStop = receiveData;

    /* start up the components */
    AuMakeElementState(&states[0], priv->flow, AuElementAll, AuStateStart);
    AuSetElementStates(aud, 1, states, ret_status);

    if (ret_flow)
	*ret_flow = priv->flow;

    if (ret_mult_elem)
	*ret_mult_elem = -1;	/* XXX - not implemented yet */

    return handler;
}

#define	VOL(volume)		((1 << 16) * (volume) / 100)

/* ARGSUSED */
static void
sync_play_cb(aud, handler, ev, data)
AuServer       *aud;
AuEventHandlerRec *handler;
AuEvent        *ev;
AuPointer       data;
{
    int            *d = (int *) data;

    *d = 1;
}

AuBool
AuSoundPlaySynchronousFromFile(aud, fname, volume)
AuServer       *aud;
_AuConst char  *fname;
int             volume;
{
    int             d = 0;
    AuStatus        ret;
    AuEvent         ev;

    if (!AuSoundPlayFromFile(aud, fname, AuNone, VOL(volume),
			   sync_play_cb, (AuPointer) & d, (AuFlowID *) NULL,
			     (int *) NULL, (int *) NULL, &ret))
    {
	/* XXX do something with ret? */
	return AuFalse;
    }
    while (1)
    {
	AuNextEvent(aud, AuTrue, &ev);
	AuDispatchEvent(aud, &ev);
	if (d)
	    break;
    }
    return AuTrue;
}

/* ARGSUSED */
/* record into a bucket */
AuEventHandlerRec *
AuSoundRecordToBucket(aud, bucket, device, gain, callback, callback_data,
		      mode, ret_flow, ret_mult_elem, ret_status)
AuServer       *aud;
AuBucketID      bucket;
AuDeviceID      device;
AuFixedPoint    gain;
void            (*callback) ();
AuPointer       callback_data;
int             mode;
AuFlowID       *ret_flow;
int            *ret_mult_elem;
AuStatus       *ret_status;
{
    AuElement       elements[2];
    AuElementState  states[1];
    PrivPtr         priv;
    AuBucketAttributes *ba;
    AuEventHandlerRec *handler;

    if (!(ba = AuGetBucketAttributes(aud, bucket, ret_status)))
	return NULL;

    /* alloc private data */
    if (!(priv = (PrivPtr) Aumalloc(sizeof(PrivRec))))
    {
	AuFreeBucketAttributes(aud, 1, ba);
	return NULL;
    }

    priv->loopCount = 0;
    priv->callback = callback;
    priv->callback_data = callback_data;
    priv->dataHandlerStop = NULL;
    priv->fp = NULL;

    /* create the flow */
    if ((priv->flow = AuGetScratchFlow(aud, NULL)) == AuNone)
    {
	AuFreeBucketAttributes(aud, 1, ba);
	Aufree(priv);
	return NULL;
    }

    AuMakeElementImportDevice(&elements[0], AuBucketSampleRate(ba), device,
			      AuUnlimitedSamples, 0, NULL);
    AuMakeElementExportBucket(&elements[1], 0, bucket, AuUnlimitedSamples,
			      0, 0, NULL);

    /* set up the flow */
    AuSetElements(aud, priv->flow, AuTrue, 2, elements, ret_status);

    /* set up the event handler */
    if (!(handler = AuRegisterEventHandler(aud, AuEventHandlerTypeMask |
					   AuEventHandlerIDMask,
					   AuEventTypeElementNotify,
					   priv->flow, EventHandler, priv)))
    {
	AuReleaseScratchFlow(aud, priv->flow, ret_status);
	AuFreeBucketAttributes(aud, 1, ba);
	Aufree(priv);
	return NULL;
    }

    /* start up the components */
    AuMakeElementState(&states[0], priv->flow, AuElementAll, AuStateStart);
    AuSetElementStates(aud, 1, states, ret_status);

    if (ret_flow)
	*ret_flow = priv->flow;

    if (ret_mult_elem)
	*ret_mult_elem = -1;	/* XXX - not implemented yet */

    AuFreeBucketAttributes(aud, 1, ba);
    return handler;
}

/* play from a bucket */
AuEventHandlerRec *
AuSoundPlayFromBucket(aud, bucket, device, volume, callback, callback_data,
		   count, ret_flow, ret_mult_elem, ret_mon_elem, ret_status)
AuServer       *aud;
AuBucketID      bucket;
AuDeviceID      device;
AuFixedPoint    volume;
void            (*callback) ();
AuPointer       callback_data;
int             count;
AuFlowID       *ret_flow;
int            *ret_mult_elem,
               *ret_mon_elem;
AuStatus       *ret_status;
{
    AuElement       elements[4];
    AuElementState  states[1];
    AuElementAction actions[4];
    PrivPtr         priv;
    unsigned int    i,
                    numActions = 0;
    AuBucketAttributes *ba;
    AuEventHandlerRec *handler;

    if (!(ba = AuGetBucketAttributes(aud, bucket, ret_status)))
	return NULL;

    /* alloc private data */
    if (!(priv = (PrivPtr) Aumalloc(sizeof(PrivRec))))
    {
	AuFreeBucketAttributes(aud, 1, ba);
	return NULL;
    }

    priv->loopCount = 0;
    priv->callback = callback;
    priv->callback_data = callback_data;
    priv->dataHandlerStop = NULL;
    priv->fp = NULL;

    /* if no ouput device was specified, look for an appropriate one */
    if (device == AuNone)
    {
	for (i = 0; i < AuServerNumDevices(aud); i++)
	    if ((AuDeviceKind(AuServerDevice(aud, i)) ==
		 AuComponentKindPhysicalOutput) &&
		AuDeviceNumTracks(AuServerDevice(aud, i)) ==
		AuBucketNumTracks(ba))
	    {
		device = AuDeviceIdentifier(AuServerDevice(aud, i));
		break;
	    }

	/* abort if we didn't find an appropriate device */
	if (device == AuNone)
	{
	    AuFreeBucketAttributes(aud, 1, ba);
	    Aufree(priv);
	    return NULL;
	}
    }

    /* create the flow */
    if ((priv->flow = AuGetScratchFlow(aud, NULL)) == AuNone)
    {
	AuFreeBucketAttributes(aud, 1, ba);
	Aufree(priv);
	return NULL;
    }

    if (count > 1)
    {
	AuMakeSendNotifyAction(&actions[0], AuStatePause, AuStateAny,
			       AuReasonAny);
	AuMakeSendNotifyAction(&actions[1], AuStateStop, AuStateAny,
			       AuReasonAny);
	AuMakeChangeStateAction(&actions[2], AuStateStop, AuStateAny,
				AuReasonUser, priv->flow, AuElementAll,
				AuStateStop);
	AuMakeChangeStateAction(&actions[3], AuStateStop, AuStateAny,
				AuReasonEOF, priv->flow, 0, AuStateStart);

	priv->loopCount = count;
	numActions = 4;
    }

    AuMakeElementImportBucket(&elements[0], AuBucketSampleRate(ba), bucket,
			      AuUnlimitedSamples, 0, numActions, actions);
    AuMakeElementMultiplyConstant(&elements[1], 0, volume);
    AuMakeElementExportDevice(&elements[2], 1, device, AuBucketSampleRate(ba),
			      AuUnlimitedSamples, 0, NULL);

    if (ret_mon_elem)
    {
	extern int      AuMonitorRate,
	                AuMonitorFormat;

	i = 4;
	*ret_mon_elem = 3;
	AuMakeElementExportMonitor(&elements[3], 0, AuMonitorRate,
				   AuMonitorFormat, AuBucketNumTracks(ba));
    }
    else
	i = 3;

    /* set up the flow */
    AuSetElements(aud, priv->flow, AuTrue, i, elements, ret_status);

    /* set up the event handler */
    if (!(handler = AuRegisterEventHandler(aud, AuEventHandlerIDMask,
					   0, priv->flow, EventHandler,
					   (AuPointer) priv)))
    {
	AuReleaseScratchFlow(aud, priv->flow, ret_status);
	AuFreeBucketAttributes(aud, 1, ba);
	Aufree(priv);
	return NULL;
    }

    /* start up the components */
    AuMakeElementState(&states[0], priv->flow, AuElementAll, AuStateStart);
    AuSetElementStates(aud, 1, states, ret_status);

    if (ret_flow)
	*ret_flow = priv->flow;

    if (ret_mult_elem)
	*ret_mult_elem = 1;

    AuFreeBucketAttributes(aud, 1, ba);
    return handler;
}
